/*
 * Copyright (c) 2007, Andreas Stricker <andy@knitter.ch>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *   * Redistributions of source code must retain the above copyright 
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above 
 *     copyright notice, this list of conditions and the following 
 *     disclaimer in the documentation and/or other materials provided 
 *     with the distribution.
 *   * The names of its contributors must not be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string.h> /* memset */
#include <stdlib.h>  /* malloc, free */
#include "cobj.h"
#include <assert.h>

#define ITEMSIZE(x) (sizeof(x)/sizeof(x[0]))

/* type functions */

#if 0
#define TYPE_TABLE_SIZE 1024
CObjType* type_table[TYPE_TABLE_SIZE] = {0};
#endif


void
cobjtype_initialize()
{
    cobjtype_register(&CObjBase_type);
}


CObjType*
cobjtype_register(CObjType *type)
{
#if 0
    CObjType **iter;
    unsigned int i;
#endif

    type->typeid = type;

#if 0
    for (i = 0, iter = type_table; 
        (i < ITEMSIZE(type_table)) && (NULL != *iter);
        iter++, i++)
        ;
    if (i >= ITEMSIZE(type_table))
        return 0;
    *iter = type;
#endif
    return type;
}

int
cobjtype_check_type(CObjType *type, CObjType *checktype)
{
    while (type) {
        if ((type == checktype) && checktype)
            return 1;
        type = type->parent;
    }
    return 0;
}

/* Object functions */

CObjType CObjBase_type = {
    .typeid         = 0,
    .parent         = NULL,
    .object_size    = sizeof(CObjBase),
    .type_size      = sizeof(CObjType),
    .name           = "CObjBase",

    .initialize     = NULL,
    .deallocate     = NULL,
};

CObjBase*
cobjbase_new(CObjType *type)
{
    CObjBase *obj;

    if (!type || type->typeid != type)
        return NULL;

    obj = (CObjBase*)malloc(type->object_size);
    if (!obj)
        return NULL;

    memset(obj, 0, type->object_size);
    obj->type = type;
    obj->ref = 1;       /* one owner */
    if (obj->type->initialize)
        if (!obj->type->initialize(obj)) {
            free(obj);
            return NULL;
        }

    return obj;
}

void
cobjbase_unref(CObjBase *self)
{
    assert(self);
    --self->ref;
    if (self->ref <= 0)
        cobjbase_free(self);
}

CObjBase*
cobjbase_ref(CObjBase *self)
{
    assert(self && self->type && self->type == self->type->typeid);
    ++self->ref;
    return self;
}

int 
cobjbase_free(CObjBase *self)
{
    assert(self && self->type == self->type->typeid);

    /* call deallocator if one is set */
    if (self->type->deallocate)
        self->type->deallocate(self);

    self->type = NULL;  /* kill type field. Type tests will fail afterwards */

    free(self);
    return 1;
}

